import discord
from pycord.multicog import add_to_group

from services.environment_service import EnvService
from models.check_model import Check
from models.autocomplete_model import (
    Settings_autocompleter,
    File_autocompleter,
    Translations_autocompleter,
)

ALLOWED_GUILDS = EnvService.get_allowed_guilds()


class Commands(discord.Cog, name="Commands"):
    """Cog containing all slash and context commands as one-liners"""

    def __init__(
        self,
        bot,
        usage_service,
        model,
        message_queue,
        deletion_queue,
        converser_cog,
        image_draw_cog,
        image_service_cog,
        moderations_cog,
        index_cog,
        translations_cog=None,
        search_cog=None,
        transcribe_cog=None,
    ):
        super().__init__()
        self.bot = bot
        self.usage_service = usage_service
        self.model = model
        self.message_queue = message_queue
        self.deletion_queue = deletion_queue
        self.converser_cog = converser_cog
        self.image_draw_cog = image_draw_cog
        self.image_service_cog = image_service_cog
        self.moderations_cog = moderations_cog
        self.index_cog = index_cog
        self.translations_cog = translations_cog
        self.search_cog = search_cog
        self.transcribe_cog = transcribe_cog

    # Create slash command groups
    dalle = discord.SlashCommandGroup(
        name="dalle",
        description="Dalle related commands",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_dalle_roles()],
    )
    gpt = discord.SlashCommandGroup(
        name="gpt",
        description="GPT related commands",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_gpt_roles()],
    )
    system = discord.SlashCommandGroup(
        name="system",
        description="Admin/System settings for the bot",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_admin_roles()],
    )
    mod = discord.SlashCommandGroup(
        name="mod",
        description="AI-Moderation commands for the bot",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_admin_roles()],
    )
    index = discord.SlashCommandGroup(
        name="index",
        description="Custom index commands for the bot",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_index_roles()],
    )
    transcribe = discord.SlashCommandGroup(
        name="transcribe",
        description="Transcription services using OpenAI Whisper",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_index_roles()],  # TODO new role checker for transcribe
    )

    #
    # System commands
    #

    @add_to_group("system")
    @discord.slash_command(
        name="settings",
        description="Get settings for GPT3Discord",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="parameter",
        description="The setting to change",
        required=False,
        autocomplete=Settings_autocompleter.get_settings,
    )
    @discord.option(
        name="value",
        description="The value to set the setting to",
        required=False,
        autocomplete=Settings_autocompleter.get_value,
    )
    @discord.guild_only()
    async def settings(
        self, ctx: discord.ApplicationContext, parameter: str = None, value: str = None
    ):
        await self.converser_cog.settings_command(ctx, parameter, value)

    @add_to_group("system")
    @discord.slash_command(
        name="settings-reset",
        description="Reset all settings for GPT3Discord",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    async def settings_reset(self, ctx: discord.ApplicationContext):
        await self.converser_cog.settings_reset_command(ctx)

    @add_to_group("system")
    @discord.slash_command(
        name="local-size",
        description="Get the size of the dall-e images folder that we have on the current system",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    async def local_size(self, ctx: discord.ApplicationContext):
        await self.image_draw_cog.local_size_command(ctx)

    @add_to_group("system")
    @discord.slash_command(
        name="clear-local",
        description="Clear the local dalleimages folder on system.",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    async def clear_local(self, ctx: discord.ApplicationContext):
        await self.image_draw_cog.clear_local_command(ctx)

    @add_to_group("system")
    @discord.slash_command(
        name="usage",
        description="Get usage statistics for GPT3Discord",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    async def usage(self, ctx: discord.ApplicationContext):
        await self.converser_cog.usage_command(ctx)

    @add_to_group("system")
    @discord.slash_command(
        name="set-usage",
        description="Set the current OpenAI usage (in dollars)",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="usage_amount",
        description="The current usage amount in dollars and cents (e.g 10.24)",
        type=float,
    )
    async def set_usage(self, ctx: discord.ApplicationContext, usage_amount: float):
        await self.converser_cog.set_usage_command(ctx, usage_amount)

    @add_to_group("system")
    @discord.slash_command(
        name="delete-conversation-threads",
        description="Delete all conversation threads across the bot servers.",
        guild_ids=ALLOWED_GUILDS,
    )
    async def delete_all_conversation_threads(self, ctx: discord.ApplicationContext):
        await self.converser_cog.delete_all_conversation_threads_command(ctx)

    # """
    # Moderation commands
    # """

    @add_to_group("mod")
    @discord.slash_command(
        name="test",
        description="Used to test a prompt and see what threshold values are returned by the moderations endpoint",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="prompt",
        description="The prompt to test",
        required=True,
    )
    @discord.guild_only()
    async def moderations_test(self, ctx: discord.ApplicationContext, prompt: str):
        await self.moderations_cog.moderations_test_command(ctx, prompt)

    @add_to_group("mod")
    @discord.slash_command(
        name="set",
        description="Turn the moderations service on and off",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="status",
        description="Enable or disable the moderations service for the current guild (on/off)",
        required=True,
        choices=["on", "off"],
    )
    @discord.option(
        name="alert_channel_id",
        description="The channel ID to send moderation alerts to",
        required=False,
        autocomplete=Settings_autocompleter.get_value_alert_id_channel,
    )
    @discord.guild_only()
    async def moderations(
        self, ctx: discord.ApplicationContext, status: str, alert_channel_id: str
    ):
        await self.moderations_cog.moderations_command(ctx, status, alert_channel_id)

    @add_to_group("mod")
    @discord.slash_command(
        name="config",
        description="Configure the moderations service for the current guild. Lower # = more strict",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="type",
        description="The type of moderation to configure",
        required=True,
        autocomplete=Settings_autocompleter.get_value_moderations,
    )
    @discord.option(
        name="hate",
        description="The threshold for hate speech",
        required=False,
        min_value=0,
        max_value=1,
    )
    @discord.option(
        name="hate_threatening",
        description="The threshold for hate/threatening speech",
        required=False,
        min_value=0,
        max_value=1,
    )
    @discord.option(
        name="self_harm",
        description="The threshold for self_harm speech",
        required=False,
        min_value=0,
        max_value=1,
    )
    @discord.option(
        name="sexual",
        description="The threshold for sexual speech",
        required=False,
        min_value=0,
        max_value=1,
    )
    @discord.option(
        name="sexual_minors",
        description="The threshold for sexual speech with minors in context",
        required=False,
        min_value=0,
        max_value=1,
    )
    @discord.option(
        name="violence",
        description="The threshold for violent speech",
        required=False,
        min_value=0,
        max_value=1,
    )
    @discord.option(
        name="violence_graphic",
        description="The threshold for violent and graphic speech",
        required=False,
        min_value=0,
        max_value=1,
    )
    @discord.guild_only()
    async def config(
        self,
        ctx: discord.ApplicationContext,
        type: str,
        hate: float,
        hate_threatening: float,
        self_harm: float,
        sexual: float,
        sexual_minors: float,
        violence: float,
        violence_graphic: float,
    ):
        await self.moderations_cog.config_command(
            ctx,
            type,
            hate,
            hate_threatening,
            self_harm,
            sexual,
            sexual_minors,
            violence,
            violence_graphic,
        )

    #
    # GPT commands
    #

    @add_to_group("gpt")
    @discord.slash_command(
        name="ask",
        description="Ask GPT3 something!",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="prompt", description="The prompt to send to GPT3", required=True
    )
    @discord.option(
        name="model",
        description="The model to use for the request",
        required=False,
        autocomplete=Settings_autocompleter.get_models,
    )
    @discord.option(
        name="private", description="Will only be visible to you", required=False
    )
    @discord.option(
        name="temperature",
        description="Higher values means the model will take more risks",
        required=False,
        min_value=0,
        max_value=2,
    )
    @discord.option(
        name="top_p",
        description="1 is greedy sampling, 0.1 means only considering the top 10% of probability distribution",
        required=False,
        min_value=0,
        max_value=1,
    )
    @discord.option(
        name="frequency_penalty",
        description="Decreasing the model's likelihood to repeat the same line verbatim",
        required=False,
        min_value=-2,
        max_value=2,
    )
    @discord.option(
        name="presence_penalty",
        description="Increasing the model's likelihood to talk about new topics",
        required=False,
        min_value=-2,
        max_value=2,
    )
    @discord.guild_only()
    async def ask(
        self,
        ctx: discord.ApplicationContext,
        prompt: str,
        model: str,
        private: bool,
        temperature: float,
        top_p: float,
        frequency_penalty: float,
        presence_penalty: float,
    ):
        await self.converser_cog.ask_command(
            ctx,
            prompt,
            private,
            temperature,
            top_p,
            frequency_penalty,
            presence_penalty,
            model=model,
        )

    @add_to_group("gpt")
    @discord.slash_command(
        name="edit",
        description="Ask GPT3 to edit some text!",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="instruction",
        description="How you want GPT3 to edit the text",
        required=True,
    )
    @discord.option(
        name="text",
        description="The text you want to edit, can be empty",
        required=False,
        default="",
    )
    @discord.option(
        name="private", description="Will only be visible to you", required=False
    )
    @discord.option(
        name="temperature",
        description="Higher values means the model will take more risks",
        required=False,
        input_type=float,
        min_value=0,
        max_value=2,
    )
    @discord.option(
        name="top_p",
        description="1 is greedy sampling, 0.1 means only considering the top 10% of probability distribution",
        required=False,
        input_type=float,
        min_value=0,
        max_value=1,
    )
    @discord.option(
        name="codex", description="Enable codex version", required=False, default=False
    )
    @discord.guild_only()
    async def edit(
        self,
        ctx: discord.ApplicationContext,
        instruction: str,
        text: str,
        private: bool,
        temperature: float,
        top_p: float,
        codex: bool,
    ):
        await self.converser_cog.edit_command(
            ctx, instruction, text, private, temperature, top_p, codex
        )

    @add_to_group("gpt")
    @discord.slash_command(
        name="converse",
        description="Have a conversation with GPT3",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="opener",
        description="Which sentence to start with, added after the file",
        required=False,
    )
    @discord.option(
        name="opener_file",
        description="Which file to start with, added before the opener, sets minimal starter",
        required=False,
        autocomplete=File_autocompleter.get_openers,
    )
    @discord.option(
        name="private",
        description="Converse in a private thread",
        required=False,
        default=False,
    )
    @discord.option(
        name="minimal",
        description="Use minimal starter text, saves tokens and has a more open personality",
        required=False,
        default=False,
    )
    @discord.option(
        name="model",
        description="Which model to use with the bot",
        required=False,
        default=False,
        autocomplete=Settings_autocompleter.get_converse_models,
    )
    @discord.option(
        name="temperature",
        description="Higher values means the model will take more risks",
        required=False,
        input_type=float,
        min_value=0,
        max_value=2,
    )
    @discord.option(
        name="top_p",
        description="1 is greedy sampling, 0.1 means only top 10%",
        required=False,
        input_type=float,
        min_value=0,
        max_value=1,
    )
    @discord.option(
        name="frequency_penalty",
        description="Decreasing the model's likelihood to repeat the same line verbatim",
        required=False,
        input_type=float,
        min_value=-2,
        max_value=2,
    )
    @discord.option(
        name="presence_penalty",
        description="Increasing the model's likelihood to talk about new topics",
        required=False,
        input_type=float,
        min_value=-2,
        max_value=2,
    )
    @discord.option(
        name="use_threads",
        description="Set this to false to start a channel conversation",
        required=False,
        default=True,
    )
    @discord.guild_only()
    async def converse(
        self,
        ctx: discord.ApplicationContext,
        opener: str,
        opener_file: str,
        private: bool,
        minimal: bool,
        model: str,
        temperature: float,
        top_p: float,
        frequency_penalty: float,
        presence_penalty: float,
        use_threads: bool,
    ):
        await self.converser_cog.converse_command(
            ctx,
            opener,
            opener_file,
            private,
            minimal,
            model,
            temperature,
            top_p,
            frequency_penalty,
            presence_penalty,
            use_threads=use_threads,
        )

    @add_to_group("gpt")
    @discord.slash_command(
        name="end",
        description="End a conversation with GPT3",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    async def end(self, ctx: discord.ApplicationContext):
        await self.converser_cog.end_command(ctx)

    #
    # Index commands
    #
    @add_to_group("index")
    @discord.slash_command(
        name="rename-user",
        description="Select one of your saved indexes to rename",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    @discord.option(
        name="user_index",
        description="Which user index to rename",
        required=True,
        autocomplete=File_autocompleter.get_user_indexes,
    )
    @discord.option(
        name="new_name",
        description="The new name",
        required=True,
        type=discord.SlashCommandOptionType.string,
    )
    async def rename_user_index(
        self,
        ctx: discord.ApplicationContext,
        user_index: str,
        new_name: str,
    ):
        await ctx.defer()
        await self.index_cog.rename_user_index_command(ctx, user_index, new_name)

    @add_to_group("index")
    @discord.slash_command(
        name="rename-server",
        description="Select one of your saved server indexes to rename",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    @discord.option(
        name="server_index",
        description="Which server index to rename",
        required=True,
        autocomplete=File_autocompleter.get_server_indexes,
    )
    @discord.option(
        name="new_name",
        description="The new name",
        required=True,
        type=discord.SlashCommandOptionType.string,
    )
    async def rename_server_index(
        self,
        ctx: discord.ApplicationContext,
        server_index: str,
        new_name: str,
    ):
        await ctx.defer()
        await self.index_cog.rename_server_index_command(ctx, server_index, new_name)

    @add_to_group("index")
    @discord.slash_command(
        name="rename-search",
        description="Select one of your saved search indexes to rename",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    @discord.option(
        name="search_index",
        description="Which search index to rename",
        required=True,
        autocomplete=File_autocompleter.get_user_search_indexes,
    )
    @discord.option(
        name="new_name",
        description="The new name",
        required=True,
        type=discord.SlashCommandOptionType.string,
    )
    async def rename_search_index(
        self,
        ctx: discord.ApplicationContext,
        search_index: str,
        new_name: str,
    ):
        await ctx.defer()
        await self.index_cog.rename_search_index_command(ctx, search_index, new_name)

    @add_to_group("index")
    @discord.slash_command(
        name="load",
        description="Select one of your saved indexes to query from",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    @discord.option(
        name="user_index",
        description="Which user file to load the index from",
        required=False,
        autocomplete=File_autocompleter.get_user_indexes,
    )
    @discord.option(
        name="server_index",
        description="Which server file to load the index from",
        required=False,
        autocomplete=File_autocompleter.get_server_indexes,
    )
    @discord.option(
        name="search_index",
        description="Which search index file to load the index from",
        required=False,
        autocomplete=File_autocompleter.get_user_search_indexes,
    )
    async def load_index(
        self,
        ctx: discord.ApplicationContext,
        user_index: str,
        server_index: str,
        search_index: str,
    ):
        await ctx.defer()
        await self.index_cog.load_index_command(
            ctx, user_index, server_index, search_index
        )

    @add_to_group("index")
    @discord.slash_command(
        name="add", description="Add an index to query from", guild_ids=ALLOWED_GUILDS
    )
    @discord.guild_only()
    @discord.option(
        name="file",
        description="A file to create the index from",
        required=False,
        input_type=discord.SlashCommandOptionType.attachment,
    )
    @discord.option(
        name="link",
        description="A link to a file to a webpage ",
        required=False,
        input_type=str,
    )
    async def set_file(
        self, ctx: discord.ApplicationContext, file: discord.Attachment, link: str
    ):
        await self.index_cog.set_index_command(ctx, file, link)

    @add_to_group("index")
    @discord.slash_command(
        name="recurse-link",
        description="Recursively index a link",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    @discord.option(
        name="link",
        description="A link to create the index from",
        required=True,
        input_type=discord.SlashCommandOptionType.string,
    )
    @discord.option(
        name="depth",
        description="How deep to recurse",
        required=False,
        input_type=discord.SlashCommandOptionType.integer,
        min_value=1,
        max_value=5,
    )
    async def set_recurse_link(
        self, ctx: discord.ApplicationContext, link: str, depth: int
    ):
        await self.index_cog.set_index_link_recurse_command(ctx, link, depth)

    @add_to_group("index")
    @discord.slash_command(
        name="reset",
        description="Reset (delete) all of your saved indexes",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    async def reset(self, ctx: discord.ApplicationContext):
        await self.index_cog.reset_command(ctx)

    @add_to_group("index")
    @discord.slash_command(
        name="compose",
        description="Combine multiple indexes together",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="name",
        description="The name of the new index",
        required=False,
        input_type=discord.SlashCommandOptionType.string,
    )
    @discord.guild_only()
    async def compose(self, ctx: discord.ApplicationContext, name: str):
        await self.index_cog.compose_command(ctx, name)

    @add_to_group("index")
    @discord.slash_command(
        name="add_discord",
        description="Set a index from a discord channel",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    @discord.option(
        name="channel",
        description="A channel to create the index from",
        required=False,
        input_type=discord.SlashCommandOptionType.channel,
    )
    @discord.option(
        name="message_limit",
        description="The number of messages to index",
        required=False,
        input_type=discord.SlashCommandOptionType.integer,
    )
    async def set_discord(
        self,
        ctx: discord.ApplicationContext,
        channel: discord.TextChannel,
        message_limit: int,
    ):
        await self.index_cog.set_discord_command(
            ctx, channel, message_limit=message_limit
        )

    @add_to_group("index")
    @discord.slash_command(
        name="discord_backup",
        description="Save an index made from the whole server",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_admin_roles(), Check.check_index_roles()],
    )
    @discord.option(
        name="message_limit",
        description="The number of messages to index per channel",
        required=False,
        input_type=discord.SlashCommandOptionType.integer,
    )
    @discord.guild_only()
    async def discord_backup(self, ctx: discord.ApplicationContext, message_limit: int):
        await self.index_cog.discord_backup_command(ctx, message_limit=message_limit)

    @add_to_group("index")
    @discord.slash_command(
        name="query", description="Query from your index", guild_ids=ALLOWED_GUILDS
    )
    @discord.guild_only()
    @discord.option(name="query", description="What to query the index", required=True)
    @discord.option(
        name="nodes",
        description="How many nodes should the response be queried from, only non-deep indexes",
        required=False,
        default=1,
        min_value=1,
        max_value=5,
        input_type=discord.SlashCommandOptionType.integer,
    )
    @discord.option(
        name="response_mode",
        description="Response mode, doesn't work on deep composed indexes",
        guild_ids=ALLOWED_GUILDS,
        required=False,
        default="default",
        choices=["default", "compact", "tree_summarize"],
    )
    @discord.option(
        name="child_branch_factor",
        description="Only for deep indexes, how deep to go, higher is expensive.",
        required=False,
        default=1,
        min_value=1,
        max_value=3,
        input_type=discord.SlashCommandOptionType.integer,
    )
    async def query(
        self,
        ctx: discord.ApplicationContext,
        query: str,
        nodes: int,
        response_mode: str,
        child_branch_factor: int,
    ):
        await ctx.defer()
        await self.index_cog.query_command(
            ctx, query, nodes, response_mode, child_branch_factor
        )

    #
    # DALLE commands
    #

    @add_to_group("dalle")
    @discord.slash_command(
        name="draw",
        description="Draw an image from a prompt",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(name="prompt", description="The prompt to draw from", required=True)
    async def draw(self, ctx: discord.ApplicationContext, prompt: str):
        await self.image_draw_cog.draw_command(ctx, prompt)

    @add_to_group("dalle")
    @discord.slash_command(
        name="optimize",
        description="Optimize a text prompt for DALL-E/MJ/SD image generation.",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.option(
        name="prompt", description="The text prompt to optimize.", required=True
    )
    @discord.guild_only()
    async def optimize(self, ctx: discord.ApplicationContext, prompt: str):
        await self.image_service_cog.optimize_command(ctx, prompt)

    #
    # Other commands
    #

    @discord.slash_command(
        name="private-test",
        description="Private thread for testing. Only visible to you and server admins.",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    async def private_test(self, ctx: discord.ApplicationContext):
        await self.converser_cog.private_test_command(ctx)

    @discord.slash_command(
        name="help", description="Get help for GPT3Discord", guild_ids=ALLOWED_GUILDS
    )
    @discord.guild_only()
    async def help(self, ctx: discord.ApplicationContext):
        await self.converser_cog.help_command(ctx)

    @discord.slash_command(
        name="setup",
        description="Setup your API key for use with GPT3Discord",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    async def setup(self, ctx: discord.ApplicationContext):
        await self.converser_cog.setup_command(ctx)

    #
    # Text-based context menu commands from here
    #

    @discord.message_command(
        name="Ask GPT", guild_ids=ALLOWED_GUILDS, checks=[Check.check_gpt_roles()]
    )
    async def ask_gpt_action(self, ctx, message: discord.Message):
        await self.converser_cog.ask_gpt_action(ctx, message)

    #
    # Image-based context menu commands from here
    #

    @discord.message_command(
        name="Draw", guild_ids=ALLOWED_GUILDS, checks=[Check.check_dalle_roles()]
    )
    async def draw_action(self, ctx, message: discord.Message):
        await self.image_draw_cog.draw_action(ctx, message)

    """
    Translation commands and actions
    """

    @discord.slash_command(
        name="translate",
        description="Translate text to a given language",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_translator_roles()],
    )
    @discord.option(name="text", description="The text to translate", required=True)
    @discord.option(
        name="target_language",
        description="The language to translate to",
        required=True,
        autocomplete=Translations_autocompleter.get_languages,
    )
    @discord.option(
        name="formality",
        description="Formal/Informal tone of translation",
        required=False,
        autocomplete=Translations_autocompleter.get_formality_values,
    )
    @discord.guild_only()
    async def translate(
        self,
        ctx: discord.ApplicationContext,
        text: str,
        target_language: str,
        formality: str,
    ):
        if self.translations_cog:
            await self.translations_cog.translate_command(
                ctx, text, target_language, formality
            )
        else:
            await ctx.respond(
                "Translations are disabled on this server.", ephemeral=True
            )

    @discord.slash_command(
        name="languages",
        description="View the supported languages for translation",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_translator_roles()],
    )
    @discord.guild_only()
    async def languages(self, ctx: discord.ApplicationContext):
        if self.translations_cog:
            await self.translations_cog.languages_command(ctx)
        else:
            await ctx.respond(
                "Translations are disabled on this server.", ephemeral=True
            )

    @discord.message_command(
        name="Translate",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_translator_roles()],
    )
    async def translate_action(self, ctx, message: discord.Message):
        if self.translations_cog:
            await self.translations_cog.translate_action(ctx, message)
        else:
            await ctx.respond(
                "Translations are disabled on this server.", ephemeral=True
            )

    # @discord.message_command(
    #     name="Paraphrase",
    #     guild_ids=ALLOWED_GUILDS,
    #     checks=[Check.check_gpt_roles()],
    # )
    # async def paraphrase_action(self, ctx, message: discord.Message):
    #     await self.converser_cog.paraphrase_action(ctx, message)

    @discord.message_command(
        name="Elaborate",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_gpt_roles()],
    )
    async def elaborate_action(self, ctx, message: discord.Message):
        await self.converser_cog.elaborate_action(ctx, message)

    @discord.message_command(
        name="Summarize",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_gpt_roles()],
    )
    async def summarize_action(self, ctx, message: discord.Message):
        await self.converser_cog.summarize_action(ctx, message)

    # Search slash commands
    @discord.slash_command(
        name="search",
        description="Search google alongside GPT3 for something",
        guild_ids=ALLOWED_GUILDS,
        checks=[Check.check_search_roles()],
    )
    @discord.option(name="query", description="The query to search", required=True)
    @discord.option(
        name="scope",
        description="How many top links to use for context",
        required=False,
        input_type=discord.SlashCommandOptionType.integer,
        max_value=16,
        min_value=1,
    )
    @discord.option(
        name="nodes",
        description="The higher the number, the more accurate the results, but more expensive",
        required=False,
        input_type=discord.SlashCommandOptionType.integer,
        max_value=8,
        min_value=1,
    )
    @discord.option(
        name="deep",
        description="Do a more intensive, long-running search",
        required=False,
        input_type=discord.SlashCommandOptionType.boolean,
    )
    @discord.option(
        name="response_mode",
        description="Response mode, doesn't work on deep searches",
        guild_ids=ALLOWED_GUILDS,
        required=False,
        default="default",
        choices=["default", "compact", "tree_summarize"],
    )
    @discord.guild_only()
    async def search(
        self,
        ctx: discord.ApplicationContext,
        query: str,
        scope: int,
        nodes: int,
        deep: bool,
        response_mode: str,
    ):
        await self.search_cog.search_command(
            ctx, query, scope, nodes, deep, response_mode
        )

    # Transcribe commands
    @add_to_group("transcribe")
    @discord.slash_command(
        name="file",
        description="Transcribe an audio or video file",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    @discord.option(
        name="file",
        description="A file to transcribe",
        required=True,
        input_type=discord.SlashCommandOptionType.attachment,
    )
    @discord.option(
        name="temperature",
        description="The higher the value, the riskier the model will be",
        required=False,
        input_type=discord.SlashCommandOptionType.number,
        max_value=1,
        min_value=0,
    )
    async def transcribe_file(
        self,
        ctx: discord.ApplicationContext,
        file: discord.Attachment,
        temperature: float,
    ):
        await self.transcribe_cog.transcribe_file_command(ctx, file, temperature)

    @add_to_group("transcribe")
    @discord.slash_command(
        name="link",
        description="Transcribe a file link or youtube link",
        guild_ids=ALLOWED_GUILDS,
    )
    @discord.guild_only()
    @discord.option(
        name="link",
        description="A link to transcribe",
        required=True,
        input_type=discord.SlashCommandOptionType.string,
    )
    @discord.option(
        name="temperature",
        description="The higher the value, the riskier the model will be",
        required=False,
        input_type=discord.SlashCommandOptionType.number,
        max_value=1,
        min_value=0,
    )
    async def transcribe_link(
        self, ctx: discord.ApplicationContext, link: str, temperature: float
    ):
        await self.transcribe_cog.transcribe_link_command(ctx, link, temperature)



"""
Store information about a discord user, for the purposes of enabling conversations. We store a message
history, message count, and the id of the user in order to track them.
"""


class RedoUser:
    def __init__(self, prompt, instruction, message, ctx, response, codex, paginator):
        self.prompt = prompt
        self.instruction = instruction
        self.message = message
        self.ctx = ctx
        self.response = response
        self.codex = codex
        self.paginator = paginator
        self.interactions = []

    def add_interaction(self, interaction):
        self.interactions.append(interaction)

    def in_interaction(self, interaction):
        return interaction in self.interactions

    # Represented by user_id
    def __hash__(self):
        return hash(self.message.author.id)

    def __eq__(self, other):
        return self.message.author.id == other.message.author.id

    # repr
    def __repr__(self):
        return f"RedoUser({self.message.author.id})"


class User:
    def __init__(self, user_id):
        self.user_id = user_id
        self.history = []
        self.count = 0

    # These user objects should be accessible by ID, for example if we had a bunch of user
    # objects in a list, and we did `if 1203910293001 in user_list`, it would return True
    # if the user with that ID was in the list
    def __eq__(self, other):
        return self.user_id == other.id

    def __hash__(self):
        return hash(self.user_id)

    def __repr__(self):
        return f"User(id={self.user_id}, history={self.history})"

    def __str__(self):
        return self.__repr__()


class Thread:
    def __init__(self, thread_id):
        self.thread_id = thread_id
        self.history = []
        self.count = 0
        self.model = None
        self.temperature = None
        self.top_p = None
        self.frequency_penalty = None
        self.presence_penalty = None

    def set_overrides(
        self,
        temperature=None,
        top_p=None,
        frequency_penalty=None,
        presence_penalty=None,
    ):
        self.temperature = temperature
        self.top_p = top_p
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty

    def get_overrides(self):
        return {
            "temperature": self.temperature,
            "top_p": self.top_p,
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
        }

    # These user objects should be accessible by ID, for example if we had a bunch of user
    # objects in a list, and we did `if 1203910293001 in user_list`, it would return True
    # if the user with that ID was in the list
    def __eq__(self, other):
        return self.thread_id == other.id

    def __hash__(self):
        return hash(self.thread_id)

    def __repr__(self):
        return f"Thread(id={self.thread_id}, history={self.history})"

    def __str__(self):
        return self.__repr__()


class EmbeddedConversationItem:
    def __init__(self, text, timestamp):
        self.text = text
        self.timestamp = int(timestamp)

    def __repr__(self):
        return self.text

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return self.text == other.text and self.timestamp == other.timestamp

    def __hash__(self):
        return hash(self.text) + hash(self.timestamp)

    def __lt__(self, other):
        return self.timestamp < other.timestamp

    def __gt__(self, other):
        return self.timestamp > other.timestamp

    def __le__(self, other):
        return self.timestamp <= other.timestamp

    def __ge__(self, other):
        return self.timestamp >= other.timestamp

    def __ne__(self, other):
        return not self.__eq__(other)

    # Make it such that if there is an array with these EmbeddedConversationItems, if we "".join the array, each item will
    # return the .text attribute
    def __format__(self, format_spec):
        return self.text



from pathlib import Path

import aiofiles
from transformers import GPT2TokenizerFast


class UsageService:
    def __init__(self, data_dir: Path):
        self.usage_file_path = data_dir / "usage.txt"
        # If the usage.txt file doesn't currently exist in the directory, create it and write 0.00 to it.
        if not self.usage_file_path.exists():
            with self.usage_file_path.open("w") as f:
                f.write("0.00")
                f.close()
        self.tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")

    async def get_price(
        self,
        tokens_used,
        prompt_tokens=None,
        completion_tokens=None,
        embeddings=False,
        chatgpt=False,
        gpt4=False,
    ):
        tokens_used = int(tokens_used)
        if chatgpt:
            price = (tokens_used / 1000) * 0.002
            return price
        elif gpt4:
            price = (prompt_tokens / 1000) * 0.03 + (completion_tokens / 1000) * 0.06
            return price
        elif not embeddings:
            price = (tokens_used / 1000) * 0.02
        else:
            price = (tokens_used / 1000) * 0.0004
        return price

    async def update_usage(
        self,
        tokens_used,
        prompt_tokens=None,
        completion_tokens=None,
        embeddings=False,
        chatgpt=False,
        gpt4=False,
    ):
        tokens_used = int(tokens_used)
        if chatgpt:
            price = (tokens_used / 1000) * 0.002
        elif gpt4:
            price = (prompt_tokens / 1000) * 0.03 + (completion_tokens / 1000) * 0.06
        elif not embeddings:
            price = (tokens_used / 1000) * 0.02
        else:
            price = (tokens_used / 1000) * 0.0004
        usage = await self.get_usage()
        print(
            f"Cost -> Old: {str(usage)} | New: {str(usage + float(price))}, used {str(float(price))} credits"
        )
        # Do the same as above but with aiofiles
        async with aiofiles.open(self.usage_file_path, "w") as f:
            await f.write(str(usage + float(price)))
            await f.close()

    async def set_usage(self, usage):
        async with aiofiles.open(self.usage_file_path, "w") as f:
            await f.write(str(usage))
            await f.close()

    async def get_usage(self):
        async with aiofiles.open(self.usage_file_path, "r") as f:
            usage = float((await f.read()).strip())
            await f.close()
        return usage

    def count_tokens(self, text):
        res = self.tokenizer(text)["input_ids"]
        return len(res)

    async def update_usage_image(self, image_size):
        # 1024×1024    $0.020 / image
        # 512×512    $0.018 / image
        # 256×256    $0.016 / image

        if image_size == "1024x1024":
            price = 0.02
        elif image_size == "512x512":
            price = 0.018
        elif image_size == "256x256":
            price = 0.016
        else:
            raise ValueError("Invalid image size")

        usage = await self.get_usage()

        async with aiofiles.open(self.usage_file_path, "w") as f:
            await f.write(str(usage + float(price)))
            await f.close()

    @staticmethod
    def count_tokens_static(text):
        tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")
        res = tokenizer(text)["input_ids"]
        return len(res)



import asyncio
import traceback
from functools import partial
from pathlib import Path

import aiohttp
import discord
from discord.ext import pages
from pytube import YouTube

from models.deepl_model import TranslationModel
from models.embed_statics_model import EmbedStatics
from services.environment_service import EnvService
from services.text_service import TextService

ALLOWED_GUILDS = EnvService.get_allowed_guilds()
USER_INPUT_API_KEYS = EnvService.get_user_input_api_keys()
USER_KEY_DB = EnvService.get_api_db()


class TranscribeService(discord.Cog, name="TranscribeService"):
    """Cog containing translation commands and retrieval of transcribe services"""

    def __init__(
        self,
        bot,
        model,
        usage_service,
    ):
        super().__init__()
        self.bot = bot
        self.model = model
        self.usage_service = usage_service
        # Make the "audiotemp" folder if it doesn't exist, using pathlib
        Path("audiotemp").mkdir(parents=True, exist_ok=True)

    async def transcribe_link_command(
        self, ctx: discord.ApplicationContext, link: str, temperature: float
    ):
        # Check if this discord file is an instance of mp3, mp4, mpeg, mpga, m4a, wav, or webm.
        await ctx.defer()

        user_api_key = None
        if USER_INPUT_API_KEYS:
            user_api_key = await TextService.get_user_api_key(
                ctx.user.id, ctx, USER_KEY_DB
            )
            if not user_api_key:
                return

        if "youtube" in link:
            # We need to download the youtube video and save it to a temporary file
            yt = YouTube(link)

            # Delete audiotemp/{str(ctx.user.id)}temp.mp3 if it already exists
            if Path("audiotemp/{}temp.mp3".format(str(ctx.user.id))).exists():
                Path("audiotemp/{}temp.mp3".format(str(ctx.user.id))).unlink()
            print("before call")
            try:
                file_path = await asyncio.get_running_loop().run_in_executor(
                    None,
                    partial(
                        yt.streams.filter().first().download,
                        output_path="audiotemp",
                        filename="{}temp".format(str(ctx.user.id)),
                    ),
                )
            except Exception as e:
                traceback.print_exc()
                await ctx.respond(
                    "Failed to download youtube video. Please try again later. "
                    + str(e)
                )
                return

            print("after call the file path was" + file_path)
        else:
            await ctx.respond(
                "Please upload a valid youtube link. Other links are not implemented yet"
            )
            return

        # Load the file object from the file_path
        file = discord.File(file_path)

        response_message = await ctx.respond(
            embed=EmbedStatics.build_transcribe_progress_embed()
        )

        try:
            response = await self.model.send_transcription_request(
                file, temperature, user_api_key
            )
            print(response)

            if len(response) > 4080:
                # Chunk the response into 2048 character chunks, each an embed page
                chunks = [response[i : i + 2048] for i in range(0, len(response), 2048)]
                embed_pages = []
                for chunk in chunks:
                    embed_pages.append(
                        discord.Embed(
                            title="Transcription Page {}".format(len(embed_pages) + 1),
                            description=chunk,
                        )
                    )

                paginator = pages.Paginator(
                    pages=embed_pages,
                    timeout=None,
                    author_check=False,
                )

                await paginator.respond(ctx.interaction)
                await response_message.delete()
                return

            await response_message.edit(
                embed=EmbedStatics.build_transcribe_success_embed(response)
            )
        except Exception as e:
            await response_message.edit(
                embed=EmbedStatics.build_transcribe_failed_embed(str(e))
            )

    async def transcribe_file_command(
        self,
        ctx: discord.ApplicationContext,
        file: discord.Attachment,
        temperature: float,
    ):
        # Check if this discord file is an instance of mp3, mp4, mpeg, mpga, m4a, wav, or webm.
        await ctx.defer()

        user_api_key = None
        if USER_INPUT_API_KEYS:
            user_api_key = await TextService.get_user_api_key(
                ctx.user.id, ctx, USER_KEY_DB
            )
            if not user_api_key:
                return

        if not file.filename.endswith(
            (".mp3", ".mp4", ".mpeg", ".mpga", ".m4a", ".wav", ".webm")
        ):
            await ctx.respond("Please upload a valid audio/video file.")
            return

        # Also check the file metadata in case it is actually an audio/video file but with a weird ending
        if not file.content_type.startswith(("audio/", "video/")):
            await ctx.respond("Please upload a valid audio/video file.")
            return

        response_message = await ctx.respond(
            embed=EmbedStatics.build_transcribe_progress_embed()
        )

        try:
            response = await self.model.send_transcription_request(
                file, temperature, user_api_key
            )

            if len(response) > 4080:
                # Chunk the response into 2048 character chunks, each an embed page
                chunks = [response[i : i + 2048] for i in range(0, len(response), 2048)]
                embed_pages = []
                for chunk in chunks:
                    embed_pages.append(
                        discord.Embed(
                            title="Transcription Page {}".format(len(embed_pages) + 1),
                            description=chunk,
                        )
                    )

                paginator = pages.Paginator(
                    pages=embed_pages,
                    timeout=None,
                    author_check=False,
                )

                await paginator.respond(ctx.interaction)

            await response_message.edit(
                embed=EmbedStatics.build_transcribe_success_embed(response)
            )
        except Exception as e:
            traceback.print_exc()
            await response_message.edit(
                embed=EmbedStatics.build_transcribe_failed_embed(str(e))
            )



import discord

from services.environment_service import EnvService
from typing import Callable

ADMIN_ROLES = EnvService.get_admin_roles()
DALLE_ROLES = EnvService.get_dalle_roles()
GPT_ROLES = EnvService.get_gpt_roles()
INDEX_ROLES = EnvService.get_index_roles()
TRANSLATOR_ROLES = EnvService.get_translator_roles()
SEARCH_ROLES = EnvService.get_search_roles()
ALLOWED_GUILDS = EnvService.get_allowed_guilds()


class Check:
    @staticmethod
    def check_admin_roles() -> Callable:
        async def inner(ctx: discord.ApplicationContext):
            if ADMIN_ROLES == [None]:
                return True

            if not any(role.name.lower() in ADMIN_ROLES for role in ctx.user.roles):
                await ctx.defer(ephemeral=True)
                await ctx.respond(
                    f"You don't have permission, list of roles is {ADMIN_ROLES}",
                    ephemeral=True,
                    delete_after=10,
                )
                return False
            return True

        return inner

    @staticmethod
    def check_dalle_roles() -> Callable:
        async def inner(ctx: discord.ApplicationContext):
            if DALLE_ROLES == [None]:
                return True
            if not any(role.name.lower() in DALLE_ROLES for role in ctx.user.roles):
                await ctx.defer(ephemeral=True)
                await ctx.respond(
                    f"You don't have permission, list of roles is {DALLE_ROLES}",
                    ephemeral=True,
                    delete_after=10,
                )
                return False
            return True

        return inner

    @staticmethod
    def check_gpt_roles() -> Callable:
        async def inner(ctx: discord.ApplicationContext):
            if GPT_ROLES == [None]:
                return True
            if not any(role.name.lower() in GPT_ROLES for role in ctx.user.roles):
                await ctx.defer(ephemeral=True)
                await ctx.respond(
                    f"You don't have permission, list of roles is {GPT_ROLES}",
                    ephemeral=True,
                    delete_after=10,
                )
                return False
            return True

        return inner

    @staticmethod
    def check_index_roles() -> Callable:
        async def inner(ctx: discord.ApplicationContext):
            if INDEX_ROLES == [None]:
                return True
            if not any(role.name.lower() in INDEX_ROLES for role in ctx.user.roles):
                await ctx.defer(ephemeral=True)
                await ctx.respond(
                    f"You don't have permission, list of roles is {INDEX_ROLES}",
                    ephemeral=True,
                    delete_after=10,
                )
                return False
            return True

        return inner

    @staticmethod
    def check_translator_roles() -> Callable:
        async def inner(ctx: discord.ApplicationContext):
            if TRANSLATOR_ROLES == [None]:
                return True
            if not any(
                role.name.lower() in TRANSLATOR_ROLES for role in ctx.user.roles
            ):
                await ctx.defer(ephemeral=True)
                await ctx.respond(
                    f"You don't have permission, list of roles is {TRANSLATOR_ROLES}",
                    ephemeral=True,
                    delete_after=10,
                )
                return False
            return True

        return inner

    @staticmethod
    def check_search_roles() -> Callable:
        async def inner(ctx: discord.ApplicationContext):
            if SEARCH_ROLES == [None]:
                return True
            if not any(role.name.lower() in SEARCH_ROLES for role in ctx.user.roles):
                await ctx.defer(ephemeral=True)
                await ctx.respond(
                    f"You don't have permission, list of roles is {SEARCH_ROLES}",
                    ephemeral=True,
                    delete_after=10,
                )
                return False
            return True

        return inner
